;;; DO NOT MODIFY THIS FILE
(if (featurep 'build-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "build/_pkg.el")

(package-provide 'build :version 1.01 :type 'regular)

;;;***

;;;### (autoloads (build build-version) "build" "build/build.el")

(autoload 'build-version "build" "\
Show the version number of `build' in the minibuffer.
If optional argument HERE is non-nil, insert info at point." t nil)

(autoload 'build "build" "\
Creates a widget-based interface to build a beta/release version of
XEmacs.  All aspects of fetching tarballs, configuring, making and
reporting can be customized and executed from the newly created buffer
*Build*." t nil)

;;;***

(provide 'build-autoloads)
